package org.cometd.client;

import java.util.HashMap;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicReference;

import junit.framework.TestCase;
import org.cometd.Bayeux;
import org.cometd.Client;
import org.cometd.Message;
import org.cometd.MessageListener;
import org.cometd.server.continuation.ContinuationCometdServlet;
import org.eclipse.jetty.client.HttpClient;
import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.server.handler.HandlerCollection;
import org.eclipse.jetty.server.nio.SelectChannelConnector;
import org.eclipse.jetty.servlet.ServletContextHandler;
import org.eclipse.jetty.servlet.ServletHolder;

public class PublishDoesNotLeakClientIdTest extends TestCase
{
    private Server server;
    private String cometdURL;
    private HttpClient httpClient;

    @Override
    protected void setUp() throws Exception
    {
        server = new Server();
        SelectChannelConnector connector = new SelectChannelConnector();
        server.addConnector(connector);

        HandlerCollection handlers = new HandlerCollection();
        server.setHandler(handlers);

        String contextPath = "/cometd";
        ServletContextHandler context = new ServletContextHandler(handlers, contextPath, ServletContextHandler.SESSIONS);

        // Setup comet servlet
        ContinuationCometdServlet cometdServlet = new ContinuationCometdServlet();
        ServletHolder cometdServletHolder = new ServletHolder(cometdServlet);
        cometdServletHolder.setInitParameter("logLevel", "3");
        String cometdServletPath = "/cometd";
        context.addServlet(cometdServletHolder, cometdServletPath + "/*");

        server.start();

        int port = connector.getLocalPort();
        String contextURL = "http://localhost:" + port + contextPath;
        cometdURL = contextURL + cometdServletPath;

        httpClient = new HttpClient();
        httpClient.start();
    }

    @Override
    protected void tearDown() throws Exception
    {
        httpClient.stop();

        server.stop();
        server.join();
    }

    public void testPublishDoesNotLeakClientId() throws Exception
    {
        final CountDownLatch connect1 = new CountDownLatch(1);
        BayeuxClient client1 = new BayeuxClient(httpClient, cometdURL)
        {
            @Override
            protected void metaConnect(boolean success, Message message)
            {
                if (success)
                    connect1.countDown();
            }
        };
        client1.start();
        try
        {
            assertTrue(connect1.await(1000, TimeUnit.MILLISECONDS));

            final CountDownLatch connect2 = new CountDownLatch(1);
            BayeuxClient client2 = new BayeuxClient(httpClient, cometdURL)
            {
                @Override
                protected void metaConnect(boolean success, Message message)
                {
                    if (success)
                        connect2.countDown();
                }
            };
            client2.start();
            try
            {
                assertTrue(connect2.await(1000, TimeUnit.MILLISECONDS));

                assertFalse(client1.getId().equals(client2.getId()));

                final String channel = "/test";
                final CountDownLatch subscribe = new CountDownLatch(1);
                final CountDownLatch latch = new CountDownLatch(1);
                final AtomicReference<Message> messageRef = new AtomicReference<Message>();
                client1.addListener(new MessageListener()
                {
                    public void deliver(Client client, Client client1, Message message)
                    {
                        if (Bayeux.META_SUBSCRIBE.equals(message.getChannel()))
                        {
                            subscribe.countDown();
                        }
                        else if (channel.equals(message.getChannel()))
                        {
                            messageRef.set(message);
                            latch.countDown();
                        }
                    }
                });
                client1.subscribe(channel);
                assertTrue(subscribe.await(1000, TimeUnit.MILLISECONDS));

                client2.publish(channel, new HashMap(), null);

                assertTrue(latch.await(1000, TimeUnit.MILLISECONDS));
                assertNull(messageRef.get().getClientId());
            }
            finally
            {
                client2.stop();
            }
        }
        finally
        {
            client1.stop();
        }
    }
}
