package org.cometd.client;

import java.io.IOException;
import java.util.HashMap;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;

import junit.framework.TestCase;
import org.cometd.Message;
import org.cometd.server.MessageImpl;
import org.cometd.server.continuation.ContinuationCometdServlet;
import org.eclipse.jetty.client.HttpClient;
import org.eclipse.jetty.client.HttpExchange;
import org.eclipse.jetty.server.Connector;
import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.server.nio.SelectChannelConnector;
import org.eclipse.jetty.servlet.ServletContextHandler;
import org.eclipse.jetty.servlet.ServletHolder;
import org.eclipse.jetty.util.log.Log;

public class SimulatedNetworkFailureTest extends TestCase
{
    private Server server;
    private String cometdURL;
    private HttpClient httpClient;
    private long timeout = 10000;

    @Override
    protected void setUp() throws Exception
    {
        server = new Server();

        Connector connector = new SelectChannelConnector();
        server.addConnector(connector);

        String contextPath = "";
        ServletContextHandler context = new ServletContextHandler(server, contextPath);

        // Cometd servlet
        ContinuationCometdServlet cometdServlet = new ContinuationCometdServlet();
        ServletHolder cometdServletHolder = new ServletHolder(cometdServlet);
        cometdServletHolder.setInitParameter("timeout", String.valueOf(timeout));
        cometdServletHolder.setInitParameter("logLevel", "2");

        String servletPath = "/cometd";
        context.addServlet(cometdServletHolder, servletPath + "/*");

        server.start();

        cometdURL = "http://localhost:" + connector.getLocalPort() + contextPath + servletPath;

        httpClient = new HttpClient();
        httpClient.start();
    }

    @Override
    protected void tearDown() throws Exception
    {
        httpClient.stop();

        server.stop();
        server.join();
    }

    public void testShortNetworkFailure() throws Exception
    {
        final CountDownLatch latch = new CountDownLatch(2);
        final AtomicReference<CountDownLatch> publishLatch = new AtomicReference<CountDownLatch>();
        final AtomicBoolean connected = new AtomicBoolean(false);

        TestBayeuxClient client = new TestBayeuxClient(httpClient, cometdURL)
        {
            @Override
            protected void send(HttpExchange exchange) throws IOException
            {
                super.send(exchange);
                if (exchange instanceof Connect)
                    latch.countDown();
            }

            @Override
            protected void metaConnect(boolean success, Message message)
            {
                super.metaConnect(success, message);

                boolean wasConnected = connected.get();
                connected.set(success);

                if (!wasConnected && connected.get())
                    Log.info("BayeuxClient connected");
                else if (wasConnected && !connected.get())
                    Log.info("BayeuxClient unconnected");
            }

            @Override
            protected void metaPublishFail(Throwable e, Message[] messages)
            {
                super.metaPublishFail(e, messages);
                publishLatch.get().countDown();
            }
        };
        client.start();
        assertTrue(latch.await(1000, TimeUnit.MILLISECONDS));

        long networkDown = 5000;
        client.setNetworkDown(networkDown);

        // Wait for a while then publish
        Thread.sleep(1000);
        publishLatch.set(new CountDownLatch(1));
        client.publish("/test", new HashMap(), null);
        assertTrue(publishLatch.get().await(1000, TimeUnit.MILLISECONDS));

        // Wait for the connect to return
        Thread.sleep(timeout);

        // Now we are simulating the network is down

        // Another publish, it must fail
        publishLatch.set(new CountDownLatch(1));
        client.publish("/test", new HashMap(), null);
        assertTrue(publishLatch.get().await(1000, TimeUnit.MILLISECONDS));

        Thread.sleep(networkDown);

        // Now another connect has been sent (the delayed one)
        Thread.sleep(timeout);
        assertTrue(connected.get());

        // We should be able to publish now
        publishLatch.set(new CountDownLatch(1));
        client.publish("/test", new HashMap(), null);
        assertFalse(publishLatch.get().await(1000, TimeUnit.MILLISECONDS));

        client.disconnect();
        client.stop();
    }

    public void testLongNetworkFailure() throws Exception
    {
        final CountDownLatch latch = new CountDownLatch(2);
        final CountDownLatch handshakeLatch = new CountDownLatch(2);
        final AtomicReference<CountDownLatch> publishLatch = new AtomicReference<CountDownLatch>();
        final AtomicBoolean connected = new AtomicBoolean(false);

        TestBayeuxClient client = new TestBayeuxClient(httpClient, cometdURL)
        {
            @Override
            protected void send(HttpExchange exchange) throws IOException
            {
                super.send(exchange);
                if (exchange instanceof Connect)
                    latch.countDown();
            }

            @Override
            protected void metaHandshake(boolean success, boolean reestablish, Message message)
            {
                super.metaHandshake(success, reestablish, message);
                handshakeLatch.countDown();
            }

            @Override
            protected void metaConnect(boolean success, Message message)
            {
                super.metaConnect(success, message);

                boolean wasConnected = connected.get();
                connected.set(success);

                if (!wasConnected && connected.get())
                    Log.info("BayeuxClient connected");
                else if (wasConnected && !connected.get())
                    Log.info("BayeuxClient unconnected");
            }

            @Override
            protected void metaPublishFail(Throwable e, Message[] messages)
            {
                super.metaPublishFail(e, messages);
                publishLatch.get().countDown();
            }
        };
        client.start();
        assertTrue(latch.await(1000, TimeUnit.MILLISECONDS));

        long networkDown = timeout + 1000;
        client.setNetworkDown(networkDown);

        // Wait for a while then publish
        Thread.sleep(1000);
        publishLatch.set(new CountDownLatch(1));
        client.publish("/test", new HashMap(), null);
        assertTrue(publishLatch.get().await(1000, TimeUnit.MILLISECONDS));

        // Wait for the connect to return
        Thread.sleep(timeout);

        // Now we are simulating the network is down

        // Another publish, it must fail
        publishLatch.set(new CountDownLatch(1));
        client.publish("/test", new HashMap(), null);
        assertTrue(publishLatch.get().await(1000, TimeUnit.MILLISECONDS));

        Thread.sleep(networkDown);

        // Now another connect has been sent (the delayed one)
        // but the server expired the client, so we handshake again
        assertTrue(handshakeLatch.await(1000, TimeUnit.MILLISECONDS));

        // We should be able to publish now
        publishLatch.set(new CountDownLatch(1));
        client.publish("/test", new HashMap(), null);
        assertFalse(publishLatch.get().await(1000, TimeUnit.MILLISECONDS));

        client.disconnect();
        client.stop();
    }

    private class TestBayeuxClient extends BayeuxClient
    {
        private long networkDown;

        private TestBayeuxClient(HttpClient client, String url)
        {
            super(client, url);
        }

        public void setNetworkDown(long time)
        {
            this.networkDown = time;
        }

        @Override
        protected boolean send(Exchange exchange, boolean backoff)
        {
            if (networkDown > 0)
            {
                if (exchange instanceof Connect)
                {
                    backoff = true;
                    while (exchange.getBackoff() < networkDown)
                        exchange.incBackoff();

                    metaConnect(false, new MessageImpl());
                }
            }
            return super.send(exchange, backoff);
        }

        @Override
        public void publish(String toChannel, Object data, String msgId)
        {
            if (networkDown > 0)
                metaPublishFail(new IOException("network down"), new Message[0]);
            else
                super.publish(toChannel, data, msgId);
        }

        @Override
        public void subscribe(String toChannel)
        {
            if (networkDown > 0)
                metaPublishFail(new IOException("network down"), new Message[0]);
            else
                super.subscribe(toChannel);
        }

        @Override
        public void unsubscribe(String toChannel)
        {
            if (networkDown > 0)
                metaPublishFail(new IOException("network down"), new Message[0]);
            else
                super.unsubscribe(toChannel);
        }

        @Override
        protected void send(HttpExchange exchange) throws IOException
        {
            Log.info(exchange + "[backoff=" + ((Exchange)exchange).getBackoff() + "]");
            if (networkDown > 0)
            {
                if (exchange instanceof Connect)
                {
                    if (((Connect)exchange).getBackoff() > 0)
                    {
                        networkDown = 0;
                        Log.info("Resetting network down");
                    }
                }
            }
            super.send(exchange);
        }
    }
}
